from AccessControl import ClassSecurityInfo
from Products.Archetypes.public import *
from Products.CMFCore.utils import _checkPermission

from xml.dom import minidom
import random
import copy
import re
import os
from StringIO import StringIO
import Globals

try:
    from reportlab.pdfgen import canvas
    from reportlab.lib.pagesizes import A4
    from reportlab.graphics.shapes import Drawing
    from reportlab.graphics.charts.piecharts import Pie
    from reportlab.graphics.charts.barcharts import VerticalBarChart

    from reportlab.pdfbase import pdfmetrics
    from reportlab.pdfbase.ttfonts import TTFont
    pdfmetrics.registerFont(TTFont('Serif', 'luxirr.ttf'))
    pdfmetrics.registerFont(TTFont('Serif-Italic', 'luxirri.ttf'))

    pdfAvailable = True
except:
    pdfAvailable = False


from config import *
import Permissions

schema = BaseSchema + Schema((
    TextField(
        'body',
        searchable = 1,
        allowable_content_types = ('txt/xml',),
        default = '<questionnaire/>',
        widget = RichWidget(
            label_msgid = 'label_xml_body',
            label = 'XML Body',
            description_msgid = 'help_xml_body',
            description = 'Enter the XML body of the questionnaire.',
            i18n_domain = 'cmfquestionnaire',
            ),
        ),

    TextField(
        'answers',
        allowable_content_types = ('txt/xml',),
        default = '<answers/>',
        widget = TextAreaWidget(
            label_msgid = 'label_answers',
            label = 'Answers',
            description_msgid = 'help_answers',
            description = 'Enter the answers of the questionnaire.',
            i18n_domain = 'cmfquestionnaire',
            visible = 0,
            ),
        ),

    BooleanField(
        'use_ticket',
        default = 0,
        widget = BooleanWidget(
            label_msgid = 'label_use_ticket',
            label = 'Use Ticket',
            description_msgid = 'help_use_ticket',
            description = 'Do you want to use the ticket mechanism in addition to roles?',
            i18n_domain = 'cmfquestionnaire',
            ),
        ),
    ))


class Questionnaire(BaseContent):
    schema = schema

    content_icon = 'questionnaire.gif'

    security = ClassSecurityInfo()
    security.setDefaultAccess('allow')

    actions = (
          { 'id': 'view',
            'name': 'View',
            'action': 'questionnaire_view_form',
            'permissions': ( Permissions.VIEW_CONTENT_PERMISSION, )
          },
          { 'id': 'design',
            'name': 'Design',
            'action': 'questionnaire_design_form',
            'permissions': ( Permissions.EDIT_CONTENT_PERMISSION, )
          },
          { 'id': 'tickets',
            'name': 'Tickets',
            'action': 'questionnaire_tickets_form',
            'permissions': ( Permissions.ADD_CONTENT_PERMISSION, ),
            'condition': 'object/use_ticket',
          },
          { 'id': 'evaluate',
            'name': 'Evaluate',
            'action': 'questionnaire_evaluate_form',
            'permissions': ( Permissions.VIEW_RESULTS_PERMISSION, )
          },
        )

    
    leftX, topY = 50, 770
    limitX, limitY = 550, 30
    chartX, chartY = 270, 140


    def _getNextId(self, tagName):
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        max = 0
        for e in document.getElementsByTagName(tagName):
            id = int(e.getAttribute('id')[1:])
            if id > max:
                max = id
        return str(max + 1)


    def _createChild(self, parent, name, attributes = {}, text = None):
        """
        creates an element with the given name and attributes in the owner
        document of the parent node, appends a text node child if text is
        given, appends the newly created element to the children of the
        parent node and returns the newly created element
        """
        o = parent.ownerDocument
        e = o.createElement(name)
        for a in attributes:
            e.setAttribute(a, str(attributes[a]))
        if text is not None:
            t = o.createTextNode(text)
            e.appendChild(t)
        parent.appendChild(e)
        return e


    def _getElementsByTagName(self, parent, tagName):
        """
        returns the list of immediate children (with a given tag name)
        of the parent node
        """
        result = []
        for n in parent.childNodes:
            if n.tagName == tagName:
                result.append(n)
        return result


    def _getGroupById(self, parent, groupId):
        """
        returns the group element with the given id
        """
        if parent.firstChild is None:
            return None

        groups = self._getElementsByTagName(parent, 'group')
        for group in groups:
            if group.getAttribute('id') == groupId:
                return group

        return parent.firstChild # why did I do that?
    

    def _getGroupInfo(self, group):
        """
        returns a dictionary containing the information of the given
        group element
        """
        dict = {}

        dict['id'] = group.getAttribute('id')
        dict['scale'] = int(group.getAttribute('scale'))

        title = ''
        titles = self._getElementsByTagName(group, 'title')
        if titles:
            if titles[0].firstChild is not None:
                title = titles[0].firstChild.nodeValue
        dict['title'] = title

        description = ''
        descriptions = self._getElementsByTagName(group, 'description')
        if descriptions:
            if descriptions[0].firstChild is not None:
                description = descriptions[0].firstChild.nodeValue
        dict['description'] = description

        legend = ''
        legends = self._getElementsByTagName(group, 'legend')
        if legends:
            options = self._getElementsByTagName(legends[0], 'option')
            for option in options:
                if option.firstChild is not None:
                    legend = legend + option.firstChild.nodeValue + '\r\n'
        dict['legend'] = legend

        return dict


    security.declareProtected(Permissions.VIEW_CONTENT_PERMISSION,
                              'getGroups')
    def getGroups(self):
        """
        returns a list of the information of all groups
        """
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        result = []
        for group in document.childNodes:
            result.append(self._getGroupInfo(group))
        return result
   

    security.declareProtected(Permissions.VIEW_CONTENT_PERMISSION,
                              'getGroup')
    def getGroup(self, groupId):
        """
        returns the group information of the group with a given id
        """
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        group = self._getGroupById(document, groupId)
        if group is not None:
            return self._getGroupInfo(group)
        return None


    security.declareProtected(Permissions.EDIT_CONTENT_PERMISSION,
                              'addGroup')
    def addGroup(self, title, scale):
        """
        adds a group with the given title and scale to the questionnaire
        """
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        groupId = 'g' + self._getNextId('group')
        group = self._createChild(document, 'group',
                                  attributes = { 'id': groupId,
                                                 'scale': str(scale) })
        self._createChild(group, 'title', text = title)
        self.body = document.toxml()
        return groupId


    security.declareProtected(Permissions.EDIT_CONTENT_PERMISSION,
                              'updateGroup')
    def updateGroup(self, groupId, title, scale, legend, description):
        """
        updates the group information of the group with the given id
        """
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        group = self._getGroupById(document, groupId)
        if group is not None:
            # create a new title element and replace the old one
            if title.strip():
                oldTitles = self._getElementsByTagName(group, 'title')
                for oldTitle in oldTitles:
                    group.removeChild(oldTitle)
                self._createChild(group, 'title', text = title)

            # update the scale attribute
            group.setAttribute('scale', str(scale))

            # create a new legend element and replace the old one
            oldLegends =  self._getElementsByTagName(group, 'legend')
            for oldLegend in oldLegends:
                group.removeChild(oldLegend)
            if legend.strip():
                newLegend = self._createChild(group, 'legend')
                options = legend.split('\r\n')
                i = 0
                for option in options:
                    i = i + 1
                    if i > scale:
                        break
                    if option.strip():
                        self._createChild(newLegend, 'option',
                                          attributes = { 'value': str(i), },
                                          text = option)

            # create a new description element and replace the old one
            oldDescriptions = self._getElementsByTagName(group,
                                                        'description')
            for oldDescription in oldDescriptions:
                group.removeChild(oldDescription)
            if description.strip():
                self._createChild(group, 'description',
                                 text = description)

            self.body = document.toxml()


    security.declareProtected(Permissions.EDIT_CONTENT_PERMISSION,
                              'deleteGroups')
    def deleteGroups(self, groupIds):
        """
        deletes the groups with the given ids from the questionnaire
        """
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        for groupId in groupIds:
            group = self._getGroupById(document, groupId)
            if group is not None:
                parent = group.parentNode
                parent.removeChild(group)
        self.body = document.toxml()


    def _getQuestionById(self, parent, questionId):
        questions = parent.getElementsByTagName('question')
        for question in questions:
            if question.getAttribute('id') == questionId:
                return question
        return None


    def _getQuestionInfo(self, question):
        dict = {}
        dict['id'] = question.getAttribute('id')
        # assert that the question element has a text
        text = self._getElementsByTagName(question, 'text')[0]
        dict['text'] = text.firstChild.nodeValue
        return dict


    security.declareProtected(Permissions.VIEW_CONTENT_PERMISSION,
                              'getQuestions')
    def getQuestions(self, groupId):
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        result = []
        group = self._getGroupById(document, groupId)
        if group is not None:
            questions = self._getElementsByTagName(group, 'question')
            for question in questions:
                result.append(self._getQuestionInfo(question))
        return result


    security.declareProtected(Permissions.EDIT_CONTENT_PERMISSION,
                              'addQuestion')
    def addQuestion(self, groupId, question):
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        group = self._getGroupById(document, groupId)
        if group is not None:
            questionId = 'q' + self._getNextId('question')
            qe = self._createChild(group, 'question',
                                   attributes = { 'id': questionId, })
            self._createChild(qe, 'text', text = question)
            self.body = document.toxml()


    security.declareProtected(Permissions.EDIT_CONTENT_PERMISSION,
                              'deleteQuestions')
    def deleteQuestions(self, groupId, questionIds):
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        group = self._getGroupById(document, groupId)
        if group is not None:
            for questionId in questionIds:
                question = self._getQuestionById(group, questionId)
                if question is not None:
                    group.removeChild(question)
            self.body = document.toxml()


    def _addTicket(self, value = ''):
        if value:
            if not hasattr(self, 'tickets'):
                self.tickets = []
            if not value in self.tickets and \
               not value in self.used_tickets:
                self.tickets.append(value)


    def _createRandomTicket(self, size = 10):
        ticket = ''
        for i in range(size):
            digit = random.randrange(0, 10)
            ticket = ticket + str(digit)
        return ticket


    security.declareProtected(Permissions.VIEW_CONTENT_PERMISSION,
                              'checkTicket')
    def checkTicket(self, ticket):
        if not self.use_ticket and not ticket:
            ticket = self.portal_membership.getAuthenticatedMember().getId()
            if not _checkPermission(Permissions.FILLOUT_PERMISSION, self):
                return False
        if hasattr(self, 'used_tickets'):
            if ticket in self.used_tickets:
                return False
        if hasattr(self, 'tickets') and self.use_ticket:
            return ticket in self.tickets
        return True


    security.declareProtected(Permissions.VIEW_RESULTS_PERMISSION,
                              'getTickets')
    def getTickets(self):
        if hasattr(self, 'tickets'):
            return self.tickets


    security.declareProtected(Permissions.VIEW_RESULTS_PERMISSION,
                              'getUsedTickets')
    def getUsedTickets(self):
        if hasattr(self, 'used_tickets'):
            return self.used_tickets


    security.declareProtected(Permissions.VIEW_CONTENT_PERMISSION,
                              'getSubmittedCount')
    def getSubmittedCount(self):
        if hasattr(self, 'used_tickets'):
            return len(self.used_tickets)
        return 0


    security.declareProtected(Permissions.ADD_CONTENT_PERMISSION,
                              'createTickets')
    def createTickets(self, count = 1):
        if count:
            for i in range(count):
                self._addTicket(self._createRandomTicket())
            self.__changed__(1)
        return count


    security.declareProtected(Permissions.ADD_CONTENT_PERMISSION,
                              'deleteAllTickets')
    def deleteAllTickets(self):
        self.tickets = []
        self.__changed__(1)


    security.declareProtected(Permissions.FILLOUT_PERMISSION,
                              'fillout')
    def fillout(self, ticket, request):
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        ansdom = minidom.parseString(str(self.answers))
        ansdoc = ansdom.documentElement
        anselem = self._createChild(ansdoc, 'form', attributes = { 'id': ticket })
        for question in document.getElementsByTagName('question'):
            questionId = question.getAttribute('id')
            self._createChild(anselem, 'question',
                              attributes = { 'id': str(questionId) },
                              text = request[questionId])
        
        if request.has_key('comment'):
            comment = request.comment.strip()
            if comment:
                self._createChild(anselem, 'comment', text = comment)
        
        if not hasattr(self, 'used_tickets'):
            self.used_tickets = []
        self.used_tickets.append(ticket)

        if hasattr(self, 'tickets'):
            if ticket in self.tickets:
                self.tickets.remove(ticket)

        self.answers = ansdoc.toxml()
    

    def _getFormById(self, parent, formId):
        for form in self._getElementsByTagName(parent, 'form'):
            if form.getAttribute('id') == formId:
                return form
        return None
            

    security.declareProtected(Permissions.VIEW_CONTENT_PERMISSION,
                              'getAnswer')
    def getAnswer(self, ticket, questionId):
        if not _checkPermission(Permissions.VIEW_RESULTS_PERMISSION, self):
            return 0
        if ticket:
            dom = minidom.parseString(str(self.answers))
            document = dom.documentElement
            form = self._getFormById(document, ticket)
            if form is not None:
                for question in self._getElementsByTagName(form, 'question'):
                    if question.getAttribute('id') == questionId:
                        return int(question.firstChild.nodeValue)
        return 0


    security.declareProtected(Permissions.VIEW_CONTENT_PERMISSION,
                              'getComment')
    def getComment(self, ticket):
        if not _checkPermission(Permissions.VIEW_RESULTS_PERMISSION, self):
            return ''
        if ticket:
            dom = minidom.parseString(str(self.answers))
            document = dom.documentElement
            form = self._getFormById(document, ticket)
            if form is not None:
                comment = self._getElementsByTagName(form, 'comment')
                if comment:
                    return comment[0].childNodes[0].nodeValue
        return ''


    security.declareProtected(Permissions.ADD_CONTENT_PERMISSION,
                              'deleteForms')
    def deleteForms(self, formIds):
        if formIds:
            dom = minidom.parseString(str(self.answers))
            document = dom.documentElement
            for formId in formIds:
                form = self._getFormById(document, formId)
                if form:
                    parent = form.parentNode
                    parent.removeChild(form)
                    self.used_tickets.remove(formId)
            self.answers = document.toxml()


    def _getAnswersByQuestionId(self, questionId):
        dom = minidom.parseString(str(self.answers))
        document = dom.documentElement
        answers = []
        for question in document.getElementsByTagName('question'):
            if question.getAttribute('id') == questionId:
                answer = int(question.firstChild.nodeValue)
                answers.append(answer)
        return answers


    def _generateSummary(self):
        dom = minidom.parseString(str(self.body))
        document = dom.documentElement
        for question in document.getElementsByTagName('question'):
            questionId = question.getAttribute('id')

            summary = {}
            answered = 0
            for answer in self._getAnswersByQuestionId(questionId):
                if answer:
                    answered += 1
                if not summary.has_key(answer):
                    summary[answer] = 1
                else:
                    summary[answer] += 1

            # compute the weighted average
            average = 0
            answer_total = 0
            for answer in summary.keys():
                answer_total += int(answer) * summary[answer]
            if answer_total > 0:
                average = round(float(answer_total) / answered, 2)

            question.setAttribute('answered', str(answered))
            question.setAttribute('average', str(average))

            # delete the old answer summaries first
            oldAnswers = self._getElementsByTagName(question, 'answer')
            for oldAnswer in oldAnswers:
                question.removeChild(oldAnswer)

            # append the new answer summaries
            for answer in summary.keys():
                attrs = { 'count': str(summary[answer]), }
                if answer:
                    attrs['percent'] = str(100 * summary[answer] / answered)
                self._createChild(question, 'answer',
                                  attributes = attrs, text = str(answer))
                
        self.body = document.toxml()


    security.declareProtected(Permissions.VIEW_RESULTS_PERMISSION,
                              'evaluate')
    def evaluate(self):
        self._generateSummary()

        dom = minidom.parseString(str(self.body))
        document = dom.documentElement

        result = []
        for question in document.getElementsByTagName('question'):
            dict = {}

            dict['id'] = question.getAttribute('id')
            dict['answered'] = int(question.getAttribute('answered'))
            dict['average'] = float(question.getAttribute('average'))

            text = self._getElementsByTagName(question, 'text')[0]
            dict['text'] = text.firstChild.nodeValue

            group = question.parentNode
            scale = int(group.getAttribute('scale'))
            summary = []
            for i in range(scale):
                summary.append((0, 0))
            answers = self._getElementsByTagName(question, 'answer')
            for answer in answers:
                answerValue = int(answer.firstChild.nodeValue)
                if answerValue > 0:
                    answerCount = int(answer.getAttribute('count'))
                    answerPerCent = int(answer.getAttribute('percent'))
                    summary[answerValue - 1] = (answerCount, answerPerCent)
            dict['answers'] = summary
            
            legends = self._getElementsByTagName(group, 'legend')
            options = {}
            if legends:
                for option in self._getElementsByTagName(legends[0], 'option'):
                    options[option.getAttribute('value')] = option.firstChild.nodeValue

            sortedOptions = []
            for i in range(len(options)):
                sortedOptions.append(options[str(i+1)])
            dict['legend'] = sortedOptions

            result.append(dict)
        return result


    security.declareProtected(Permissions.VIEW_RESULTS_PERMISSION,
                              'getComments')
    def getComments(self):
        dom = minidom.parseString(str(self.answers))
        document = dom.documentElement
        result = []
        for comment in document.getElementsByTagName('comment'):
            if comment.firstChild.nodeValue:
                dict = {}
                dict['id'] = comment.parentNode.getAttribute('id')
                dict['text'] = comment.firstChild.nodeValue 
                result.append(dict)
        return result
    
    
    security.declareProtected(Permissions.VIEW_RESULTS_PERMISSION,
                              'getPDFReport')
    def isPDFAvailable(self):
        return pdfAvailable
    
    
    security.declareProtected(Permissions.VIEW_RESULTS_PERMISSION,
                              'getPDFReport')
    def getPDFReport(self):
        pdfFileName = os.path.join(Globals.INSTANCE_HOME, 'var', '%s.pdf' % self.id)
        pdfReport = canvas.Canvas(pdfFileName, pagesize = A4)
        pdfReport.setTitle(self.title)
        pdfReport.setAuthor('CMFQuestionnaire')

        curX, curY = self._putPDFText(self.title,
                                      pdfReport, self.leftX, self.topY,
                                      fontSize = 20)

        submitCount = len(self.used_tickets)
        curX, curY = self._putPDFText('Total submissions: %d' % submitCount,
                                      pdfReport, curX, curY,
                                      font = 'Serif-Italic')

        valueStep = 1
        if submitCount > 10:
            valueStep = 5
        if submitCount > 50:
            valueStep = 10
        if submitCount > 100:
            valueStep = 50

        results = self.evaluate()
        for result in results:
            lines = 1 + (len(self.title) / 100)
            height = lines * 12 + self.chartY + 20
            if curY - height < self.limitY:
                pdfReport.showPage()
                curX, curY = self.leftX, self.topY

            pdfReport.line(self.leftX, curY, self.limitX, curY)
            curX, curY = self._putPDFText(result['text'], pdfReport, self.leftX, curY - 20)

            barChart = VerticalBarChart()
            barChart.valueAxis.valueMin = 0
            barChart.valueAxis.valueMax = submitCount + 1
            barChart.valueAxis.valueStep = valueStep

            barChart.categoryAxis.categoryNames = []
            barChart.data = []
            values = []
            i = 0
            for answer in result['answers']:
                i = i + 1
                barChart.categoryAxis.categoryNames.append(str(i))
                values.append(answer[0])
            barChart.data.append(values)

            barChart.x = 20
            barChart.y = 20
            barChart.width = 100
            barChart.height = 100
            
            pieChart = Pie()
            pieChart.labels = barChart.categoryAxis.categoryNames
            pieChart.data = []
            for answer in result['answers']:
                pieChart.data.append(answer[1])

            pieChart.x = barChart.x + barChart.width + 30
            pieChart.y = barChart.y - 10
            pieChart.width = 100
            pieChart.height = 100

            chart = Drawing(self.chartX, self.chartY)
            chart.add(barChart)
            chart.add(pieChart)

            nextY = curY - self.chartY + 10
            chart.drawOn(pdfReport, self.leftX, nextY)

            curX, curY = self._putPDFText('Answer count: %s' % result['answered'],
                                         pdfReport, self.leftX + self.chartX + 10, curY - 10,
                                         font = 'Serif-Italic', fontSize = 10)

            curX, curY = self._putPDFText('Average: %s' % result['average'],
                                         pdfReport, self.leftX + self.chartX + 10, curY - 10,
                                         font = 'Serif-Italic', fontSize = 10)
            curY = curY - 10

            legend = result['legend']
            for i in range(len(legend)):
                curX, curY = self._putPDFText('%d: %s' % (i + 1, legend[i]),
                                             pdfReport, self.leftX + self.chartX + 10, curY,
                                             font = 'Serif-Italic', fontSize = 10)
                
            curX, curY = self.leftX, nextY - 20

        pdfReport.showPage()

        curX, curY = self._putPDFText('Comments',
                                      pdfReport, self.leftX, self.topY,
                                      fontSize = 20)
        for comment in self.getComments():
            commentText = comment['text']
            # estimate a height and check whether we should start a new page
            lines = 1 + (len(commentText) / 100)
            height = lines * 12 + 20
            if curY - height < self.limitY:
                pdfReport.showPage()
                curX, curY = self.leftX, self.topY
            pdfReport.line(self.leftX, curY, self.limitX, curY)
            curX, curY = self._putPDFText(commentText, pdfReport, self.leftX, curY - 20)
            
        pdfReport.showPage()
        pdfReport.save()
        
        pdfFile = open(pdfFileName)
        pdfString = StringIO(pdfFile.read())
        pdfFile.close()
        return pdfString.read()
    
    
    def _putPDFText(self, text, pdfReport, curX, curY,
                    limitX = None, limitY = None,
                    font = 'Serif', fontSize = 12):
        if limitX == None:
            limitX = self.limitX
        if limitY == None:
            limitY = self.limitY
        tmpText = canvas.textobject.PDFTextObject(pdfReport, x = curX, y = curY)
        pdfText = canvas.textobject.PDFTextObject(pdfReport, x = curX, y = curY)
        tmpText.setFont(font, fontSize)
        pdfText.setFont(font, fontSize)
        for word in text.split(' '):
            if word.strip():
                tmpText.textOut(word + ' ')
                if tmpText.getX() > limitX:
                    pdfText.textLine('')
                    pdfText.textOut(word + ' ')
                    tmpText.textLine('')
                    tmpText.textOut(word + ' ')
                else:
                    pdfText.textOut(word + ' ')
        pdfText.textLine('')
        curX, curY = pdfText.getX(), pdfText.getY()
        pdfReport.drawText(pdfText)
        return curX, curY

    
registerType(Questionnaire)
